import 'dart:async';

import 'jsontoken.dart';
import 'streamer.dart';

class JsonTokenizer {
  Streamer stream;
  JsonToken token;
  JsonTokenizer(Streamer stream) {
    this.stream = stream;
    this.token = new JsonToken(null, JsonToken.NULL, position: 0);
  }
  Future<int> tokenize() async {
    Completer<int> completer = new Completer<int>();
    this.stream.next().then((String response) {
      if (response == null) {
        completer.complete(1);
      } else {
        response = this.tidy(response);
        response.split("").forEach((String rawToken) {
          JsonToken other =
              new JsonToken(rawToken, JsonToken.getType(rawToken));
          this.token = this.token =
              (this.token != null ? this.token : JsonToken.queue.values.last)
                  .append(other);
        });
        this.destroy();
        new Timer(new Duration(microseconds: 10), () {
          completer.complete(0);
        });
      }
    }).catchError((err) {
      completer.completeError(err);
    });
    return completer.future;
  }

  //remove completed items
  destroy() {
    List deletables = new List();
    JsonToken.queue.forEach((int key, JsonToken value) {
      if ((value.value == null || value.value == ',' || value.value == ':') &&
          value.position != this.token.position) {
        deletables.add(key);
      }
    });
    deletables.forEach((v) {
      if (JsonToken.queue.containsKey(v)) {
        JsonToken.queue.remove(v);
      }
    });
    //delete sendables
  }

  String tidy(String sequence) {
    return sequence
        .replaceAll(new RegExp(r"(\s+|)\{(\s+|)"), '{')
        .replaceAll(new RegExp(r"(\s+|)\:(\s+|)"), ':')
        .replaceAll(new RegExp(r"(\s+|),(\s+|)"), ',')
        .replaceAll(new RegExp(r"(\s+|)\}(\s+|)"), '}')
        .replaceAll(new RegExp(r"(\s+|)\[(\s+|)"), '[')
        .replaceAll(new RegExp(r"(\s+|)\](\s+|)"), ']')
        .replaceAll(new RegExp(r"[\r\n\t]+(.+[\r\n\t]+.+)[\t\r\n]+"), '');
  }
}
