import 'dart:async';
import 'dart:convert' as JSON;
import 'package:heavylist/heavylist.dart';

import 'jsonexpression.dart';
import 'jsontoken.dart';
import 'jsontokenizer.dart';
import 'streamer.dart';

class Reader {
  JsonTokenizer tokenizer;
  Function _pipe;
  Function _trailing;
  Map<JsonExpression, Function> _filters;
  Function _done;
  Function _fail;
  Duration _delay;
  Reader(Streamer stream, {int delay = 0}) {
    this._delay = new Duration(microseconds: delay);
    this._filters = new Map<JsonExpression, Function>();
    this.tokenizer = new JsonTokenizer(stream);
    new Timer(new Duration(microseconds: 0), () {
      this._run();
    });
  }
  _run() {
    this.tokenizer.tokenize().then((item) {
      if (item == 1) {
        if (this._done != null) {
          this._done();
        }
      } else {
        try {
          this._parse().then((String d) {
            new Timer(_delay, () {
              this._run();
            });
          }).catchError((err) {
            this._fail(err);
          });
        } catch (e) {
          if (this._fail != null) {
            this._fail(e);
          }
        }
      }
    }).catchError((err) {
      if (this._fail != null) {
        this._fail(err);
      }
    });
  }

  Future<String> _parse() async {
    Completer<String> completer = new Completer<String>();
    List deletes = [];
    List<List<dynamic>> callList = [];
    List keys = JsonToken.queue.keys.toList();
    int len = keys.length;
    Map<int, bool> trailing = new Map<int, bool>();
    first:
    for (int i = 0; i < len; i++) {
      int key = keys[i];
      JsonToken item = JsonToken.queue[key];
      if (item == null) {
        continue;
      }
      if (item.type == JsonToken.OBJECTSTART ||
          item.type == JsonToken.ARRAYSTART) {
        int next = i + 1;
        if (next < len) {
          int _key = keys[next];
          JsonToken _item = JsonToken.queue[_key];
          trailing[_item.parnterId] = true;
        }
      }
      if (item.type == JsonToken.OBJECTEND || item.type == JsonToken.ARRAYEND) {
        int _key = keys[i - 1];
        JsonToken _item = JsonToken.queue[_key];
        trailing[_item.parnterId] = false;
      }
      if (item.closed == true &&
          item.type == JsonToken.OBJECTSTART &&
          item.child == null) {
        List<JsonToken> stack = new List<JsonToken>();
        JsonToken start = item;
        stack.add(start);
        for (int j = i + 1; j < len; j++) {
          int key = keys[j];
          JsonToken item = JsonToken.queue[key];
          if (item.type == JsonToken.OBJECTEND &&
              item.parnterId == start.parnterId) {
            item.child = null;
            item.partner.child = null;
            stack.add(item);
            if (stack.length > 2) {
              //send data
              List obj = this._object(stack);
              List<String> ikeys = obj[0].keys.toList();
              if (this._pipe != null) {
                callList.add([this._pipe, obj[0], obj[1]]);
              }
              for (JsonExpression k in this._filters.keys) {
                if (k.and(ikeys, obj[1])) {
                  callList.add([this._filters[k], obj[0], obj[1]]);
                }
              }
              if (this._trailing != null) {
                List<int> trailKeys = trailing.keys.toList();
                int k = trailKeys[trailKeys.length - 1];
                if (trailing[k]) {
                  if (trailing[k].runtimeType == bool) {
                    callList.add([this._trailing, obj[0], obj[1]]);
                  }
                }
              }
            }
            i = j;
            continue first;
          } else {
            int next = i + 1;
            deletes.add(key);
            deletes.add(keys[next]);
            stack.add(item);
          }
        }
      } else if (item.type == JsonToken.VALUE && item.closed == true) {
        if (item.value == 'true') {
          item.value = true;
        } else if (item.value == 'false') {
          item.value = false;
        } else if (item.value == 'null') {
          item.value = null;
        }
        RegExp r = new RegExp(r'^"|"$');
        String _data = item.value.replaceAll(r, "");
        if (this._pipe != null) {
          //send data
          callList.add([this._pipe, _data, item.key]);
        }
        for (JsonExpression k in this._filters.keys) {
          if (new RegExp(k.toPath()).hasMatch(item.key)) {
            callList.add([this._filters[k], _data, item.key]);
          }
        }
        if (this._trailing != null) {
          for (int k in trailing.keys) {
            if (trailing[k]) {
              if (trailing[k].runtimeType == bool) {
                callList.add([this._trailing, _data, item.key]);
              }
            }
          }
        }
        int next = i + 1;
        deletes.add(keys[next]);
        deletes.add(key);
      }
      if (trailing.keys.length > 0 &&
          (item.type == JsonToken.VALUEPAIR && item.closed == true) &&
          i > 2) {
        int _key_1 = keys[i - 1];
        RegExp r = new RegExp(r'^"|"$');
        JsonToken _item_1 = JsonToken.queue[_key_1];
        List<String> ikeys = [_item_1.value.replaceAll(r, "")];
        if (_item_1.type == JsonToken.KEY) {
          deletes.add(_key_1);
        }
        deletes.add(key);
        var obj = {};
        obj[_item_1.value] = item.value.replaceAll(r, "");
        if (this._trailing != null) {
          callList.add([this._trailing, obj, _item_1.key]);
        }
        for (JsonExpression k in this._filters.keys) {
          if (k.and(ikeys, _item_1.key)) {
            callList.add([this._filters[k], obj, _item_1.key]);
          }
        }
      }
    }
    for (int m = 0; m < len; m++) {
      int key = keys[m];
      JsonToken item = JsonToken.queue[key];
      if (item == null) {
        continue;
      }
      if (item.parent != null) {
        if (item.parent.closed) {
          item.closed = true;
          deletes.add(key);
        } else if (item.closed == true) {
          deletes.add(key);
        }
      } else if (item.closed == true) {
        deletes.add(key);
      }
      if (item.type == JsonToken.KEY) {
        int next = m + 1;
        if (next < len) {
          JsonToken sub = JsonToken.queue[keys[next]];
          if (sub.closed == true) {
            deletes.add(key);
          }
        }
      }
    }

    HeavyList<List<dynamic>> list = new HeavyList<List<dynamic>>(callList);
    list.loop(new Duration(microseconds: Streamer.TIMEOUT),
        (List<List<dynamic>> origin) {
      list.empty();
      deletes.forEach((v) {
        if (JsonToken.queue.containsKey(v)) {
          JsonToken.queue.remove(v);
        }
      });
      completer.complete("done");
    }, (List item, Function resume) {
      try {
        var response = Function.apply(item[0], item.sublist(1, item.length));
        bool isFuture = response.runtimeType.toString().contains('Future');
        if (isFuture) {
          response.then((dynamic v) {
            resume();
          }).catchError((error) {
            completer.completeError(error);
          });
        } else {
          resume();
        }
      } catch (err) {
        completer.completeError(err);
      }
    });
    return completer.future;
  }

  Reader pipe(Future func(dynamic value, String key)) {
    this._pipe = func;
    return this;
  }

  Reader trailing(Future func(dynamic value, String key)) {
    this._trailing = func;
    return this;
  }

  Reader done(Function func) {
    this._done = func;
    return this;
  }

  Reader fail(Function func(dynamic err)) {
    this._fail = func;
    return this;
  }

  Reader filter(
      String expression, Future<dynamic> func(dynamic value, String key)) {
    this._filters[new JsonExpression(expression)] = func;
    return this;
  }

  List<dynamic> _object(List<JsonToken> arr) {
    var alen = arr.length;
    var string = '' + arr[0].value;
    if (alen > 2) {
      var key = arr[1].key;
      for (var k = 1; k < alen - 1; k++) {
        JsonToken current = arr[k];
        String v = current.value;
        if (current.type == JsonToken.KEY) {
          string += v + ':';
        } else if (current.type == JsonToken.VALUEPAIR) {
          string += v + ',';
        }
      }
      string = string.substring(0, string.length - 1) +
          arr[alen - 1].value.toString();
      return [JSON.jsonDecode(string), key];
    } else {
      return [{}, null];
    }
  }
}
